@@ -71,9 +71,15 @@ kde.balloon.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.
         }
     if (verbose) close(pb)
 
+    ## re-scale density estimate to integral 1
     delta.int <- prod(sapply(fhat$eval.points, diff)[1,]) 
     riemann.sum <- sum(fhat$estimate*delta.int)
-    fhat.estimate <- fhat$estimate/riemann.sum
+    fhat$estimate <- fhat$estimate/riemann.sum
+    
+    fhat$names <- parse.name(x)  ## add variable names
+    if (compute.cont)
+        fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)
+
     fhat$H <- h.pi^2
     
     return(fhat)
@@ -114,9 +120,10 @@ kde.sp <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.point
         H <- Hns(x=x, deriv.order=4)
     }
 
-    if (d==2) fhat <- kde.sp.2d(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont, verbose=verbose)
+    if (d==2) fhat <- kde.sp.2d(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont, verbose=verbose, pre=TRUE)
     else stop("d!=2 not yet implemented.")
 
+    fhat$names <- parse.name(x)  ## add variable names
     if (compute.cont)
         fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)
 
@@ -127,11 +134,22 @@ kde.sp <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.point
 ## Bivariate sample point variable KDE
 ######################################################################
 
-kde.sp.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE)
+kde.sp.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE, pre=TRUE)
 {
     d <- 2
     n <- nrow(x)
     
+    if (pre) 
+    { 
+        x.orig <- x
+        S12 <- diag(apply(x.orig, 2, sd))  
+        Sinv12 <- matrix.pow(S12,-1)
+        x <- pre.scale(x)
+        if (!missing(xmin)) xmin <- xmin %*% Sinv12
+        if (!missing(xmax)) xmax <- xmax %*% Sinv12
+        H <- Hns(x=x, deriv.order=4)
+    }
+        
     fhat <- kde(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont)
     fhat.pilot <- fhat
     fhat1.pilot <- kdde(x=x, deriv.order=1, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w)
@@ -150,11 +168,11 @@ kde.sp.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.point
     lambda3 <- 1/fhat.pilot^3*rowKpow(fhat2.pilot, r=2)
     lambda4 <- 1/fhat.pilot^3*rowKpow(fhat3.pilot, fhat1.pilot, r=1, s=1)
     lambda5 <- 1/fhat.pilot^2*fhat4.pilot
-    lambda <- drop((24*lambda1 - 36*lambda2 + 6*lambda3 + 8*lambda4 - 2*lambda5) %*% Sdr(d,r=4) %*% (vec(diag(d)) %x% vec(diag(d))))
+    lambda <- drop((24*lambda1 - 36*lambda2 + 6*lambda3 + 8*lambda4 - lambda5) %*% Sdr(d,r=4) %*% (vec(diag(d)) %x% vec(diag(d))))
 
     RK <- (4*pi)^(-d/2)
-    h.Ab  <- (8*d*RK*fhat.pilot^(1+d/2)/lambda^2)^(-1/(d+8))*n^(-1/(d+8))
-
+    h.Ab  <- (8*d*RK*fhat.pilot^(1+d/2)/lambda^2)^(1/(d+8))*n^(-1/(d+8))
+    
     fhat$estimate <- array(0, dim=dim(fhat$estimate))
     xmin <- sapply(fhat$eval.points, min)
     xmax <- sapply(fhat$eval.points, max)
@@ -162,12 +180,22 @@ kde.sp.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.point
     if (verbose) { pb <- txtProgressBar(max=n) }
     for (i in 1:n)
     {
-        if (verbose) setTxtProgressBar(pb, n)
-        fhat$estimate <- fhat$estimate + kde(x=matrix(x[i,], nrow=1), H=h.Ab[i]^2*diag(d), xmin=xmin, xmax=xmax, binned=binned, gridsize=dim(fhat$estimate), bgridsize=dim(fhat$estimate))$estimate
+        if (verbose) setTxtProgressBar(pb, i)
+        HAb <- h.Ab[i]^2*diag(d)
+        fhat$estimate <- fhat$estimate + kde(x=matrix(x[i,], nrow=1), H=HAb, xmin=xmin, xmax=xmax, binned=binned, gridsize=dim(fhat$estimate), bgridsize=dim(fhat$estimate))$estimate
     }
     if (verbose) close(pb)
-
+    
     fhat$estimate <- fhat$estimate/n
+    if (pre)
+    {
+        ep <- cbind(fhat$eval.points[[1]], fhat$eval.points[[2]]) %*% S12
+        fhat$eval.points[[1]] <- ep[,1]
+        fhat$eval.points[[2]] <- ep[,2]
+        fhat$estimate <- fhat$estimate/det(S12)
+        fhat$x <- x.orig
+    }
+    fhat$cont <- contourLevels(fhat, cont=1:99)
     fhat$H <- h.Ab^2
     
     return(fhat)
